import shutil
import os
from tarpan.testutils.a01_eight_schools.eight_schools import get_fit
from tarpan.cmdstanpy.analyse import save_analysis


def test_save_analysis():
    fit = get_fit()

    outdir = "tarpan/cmdstanpy/model_info/analyse_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_analysis(fit)

    # Diagnostic
    assert os.path.isfile(os.path.join(outdir, "diagnostic.txt"))

    # Check summary files
    # ---------------

    assert os.path.isfile(os.path.join(outdir, "summary.txt"))
    assert os.path.isfile(os.path.join(outdir, "summary.csv"))

    # Tree plot
    assert os.path.isfile(os.path.join(outdir, "summary.pdf"))

    # Trace plots
    # ----------

    assert os.path.isfile(os.path.join(outdir, "traceplot_01.pdf"))
    assert os.path.isfile(os.path.join(outdir, "traceplot_02.pdf"))
    assert os.path.isfile(os.path.join(outdir, "traceplot_03.pdf"))

    # Histograms of posterior distributions
    # ----------

    assert os.path.isfile(os.path.join(outdir, "histogram_01.pdf"))
    assert os.path.isfile(os.path.join(outdir, "histogram_02.pdf"))
    assert os.path.isfile(os.path.join(outdir, "histogram_03.pdf"))

    # Pair plot
    assert os.path.isfile(os.path.join(outdir, "pair_plot.pdf"))



"""Make pair plot of parameters"""

from dataclasses import dataclass
import seaborn as sns
import matplotlib.pyplot as plt
from tarpan.shared.param_names import filter_param_names
from tarpan.shared.info_path import InfoPath, get_info_path
import math


@dataclass
class PairPlotParams:
    title: str = None  # Plot's title
    color: str = "#00a6ff"
    edgecolor: str = "#00a6ff55"
    diag_edge_color: str = "#ffffff"  # Edge color for histograms on diagonal
    alpha: float = 0.15  # Transparency of the marker color
    marker_size: float = 30
    max_params: int = 4  # Maximum number of parameter to show in the plot
    max_samples: int = 717  # Maximum number of samples to show in pair plot

    """Type of diahonal plots: 'auto’, ‘hist’, ‘kde’, None"""
    diag_kind: str = 'kde'


def save_pair_plot(samples, param_names=None,
                   info_path=InfoPath(),
                   pair_plot_params=PairPlotParams()):
    """
    Make histograms for the parameters from posterior destribution.

    Parameters
    -----------

    samples : Panda's DataFrame

        Each column contains samples from posterior distribution.

    param_names : list of str

        Names of the parameters for plotting. If None, all will be plotted.
    """

    info_path = InfoPath(**info_path.__dict__)

    g = make_pair_plot(
        samples, param_names=param_names,
        pair_plot_params=pair_plot_params)

    info_path.base_name = info_path.base_name or "pair_plot"
    info_path.extension = info_path.extension or 'pdf'
    plot_path = get_info_path(info_path)
    g.savefig(plot_path, dpi=info_path.dpi)


def make_pair_plot(samples, param_names=None,
                   pair_plot_params=PairPlotParams()):
    """
    Make a pair plot for the parameters from posterior destribution.

    Parameters
    -----------

    samples : Panda's DataFrame

        Each column contains samples from posterior distribution.

    param_names : list of str

        Names of the parameters for plotting. If None, all will be plotted.

    Returns
    -------
    Seaborn's PairGrid
    """

    param_names = filter_param_names(samples.columns, param_names)

    if len(param_names) > pair_plot_params.max_params:
        print((
            f'Showing only first {pair_plot_params.max_params} '
            f'parameters out of {len(param_names)} in pair plot.'
            'Consider limiting the parameter with "param_names".'))

        param_names = param_names[:pair_plot_params.max_params]

    samples = samples[param_names]

    # Show no more than `max_samples` markers
    keep_nth = math.ceil(samples.shape[0] / pair_plot_params.max_samples)
    samples = samples[::keep_nth]

    g = sns.PairGrid(samples)

    g = g.map_upper(sns.scatterplot, s=pair_plot_params.marker_size,
                    color=pair_plot_params.color,
                    edgecolor=pair_plot_params.edgecolor,
                    alpha=pair_plot_params.alpha)

    g = g.map_lower(sns.kdeplot, color=pair_plot_params.color)
    g = g.map_diag(plt.hist, color=pair_plot_params.color,
                   edgecolor=pair_plot_params.diag_edge_color)

    return g



"""Create summary of parameter distributions: mean, std, mode etc."""

from dataclasses import dataclass, field
from tarpan.shared.info_path import InfoPath, get_info_path
import arviz as az
import numpy as np
import pandas as pd
from tabulate import tabulate
from scipy.stats import gaussian_kde
from typing import List


@dataclass
class SummaryParams:
    # List of probabilities for HPDIs (highest posterior density intervals)
    # to be shown in summary
    hpdis: List = field(default_factory=lambda: [0.6827, 0.9545])

    def hpdi_percent(self):
        """
        Returns
        -------

        float : rounded HPDI percent value, i. e. 68 for 0.6827.
        """

        return [
            int(round(fraction, 2) * 100) for fraction in self.hpdis
        ]


def save_summary(samples, info_path=InfoPath(),
                 summary_params=SummaryParams()):
    """
    Generates and saves statistical summary of the samples using mean, std, mode, hpdi.

    Parameters
    ----------

    samples : Panda's dataframe

        Each column contains samples for a parameter.

    info_path : InfoPath

        Path information for creating summaries.
    """

    df_summary, table = sample_summary(samples, params=summary_params)
    return save_summary_to_disk(df_summary, table, info_path)


def save_summary_to_disk(df_summary, txt_summary, info_path=InfoPath()):
    """
    Saves statistical summary of the samples using mean, std, mode, hpdi.

    Parameters
    ----------

    df_summary : cmdstanpy.stanfit.CmdStanMCMC

        Panda's dataframe containing the summary for all parameters.

    txt_summary : list of str

        Text of the summary table.

    info_path : InfoPath

        Path information for creating summaries.


    Returns
    --------

    Dict:

        "df" : Dataframe containing summary.

        "table" : text version of the summary.

        "path_txt": Path to txt summary file.

        "path_csv": Path to csv summary file.

    """
    info_path = InfoPath(**info_path.__dict__)
    info_path.base_name = info_path.base_name or "summary"
    info_path.extension = 'txt'
    path_to_summary_txt = get_info_path(info_path)
    info_path.extension = 'csv'
    path_to_summary_csv = get_info_path(info_path)

    with open(path_to_summary_txt, "w") as text_file:
        print(txt_summary, file=text_file)

    df_summary.to_csv(path_to_summary_csv, index_label='Name')

    return {
        "df": df_summary,
        "table": txt_summary,
        "path_txt": path_to_summary_txt,
        "path_csv": path_to_summary_csv,
    }


def get_mode(values):
    """
    Calculates mode of the gaussian distribution
    (value at which it is maximum)
    """

    kernel = gaussian_kde(values)

    # Evaluate the kernel at no more than `max_values` for performance
    max_values = 100
    take_every = int(len(values) / max_values)

    if take_every < 1:
        take_every = 1

    values = values[0::take_every].tolist()
    imax = np.argmax(kernel(values))
    return values[imax]


def sample_summary(df, extra_values=None, params=SummaryParams()):
    """
    Returns table showing statistical summary from the sample parameters:
    mean, std, mode, hpdi.

    Parameters
    ------------

    df : Panda's dataframe

        Contains parameter sample values: each column is a parameter.

    extra_values : Panda's dataframe

        Additional values to be shown for parameters. Indexes are
        parameter names, and columns contain additional values to
        be shown in summary.

    Returns
    -------
    Panda's dataframe
        Panda's dataframe containing the summary for all parameters.
    str
        text of the summary table
    """
    rows = []

    for column in df:
        values = df[column].to_numpy()
        mean = df[column].mean()
        std = df[column].std()
        mode = get_mode(df[column])

        summary_values = [column, mean, std, mode]

        for i, hpdi in enumerate(params.hpdis):
            hpdi_value = az.hpd(values, credible_interval=hpdi).tolist()

            if i == 0:
                # For the first interval, calculate upper and
                # lower uncertainties
                uncert_plus = hpdi_value[1] - mode
                uncert_minus = mode - hpdi_value[0]
                summary_values.append(uncert_plus)
                summary_values.append(uncert_minus)

            summary_values.append(hpdi_value[0])
            summary_values.append(hpdi_value[1])

        if extra_values is not None:
            # Add extra columns
            summary_values += extra_values.loc[column].values.tolist()

        rows.append(summary_values)

    headers = ['Name', 'Mean', 'Std', 'Mode', '+', '-']

    for hpdi_percent in params.hpdi_percent():
        headers.append(f'{hpdi_percent}CI-')
        headers.append(f'{hpdi_percent}CI+')

    if extra_values is not None:
        headers += extra_values.columns.values.tolist()

    formats = [".2f"] * len(headers)

    if 'N_Eff' in headers:
        formats[headers.index('N_Eff')] = ".0f"

    table = tabulate(rows, headers=headers, floatfmt=formats, tablefmt="pipe")

    df_summary = pd.DataFrame(rows, columns=headers, index=df.columns.values)
    df_summary.drop('Name', axis=1, inplace=True)
    return df_summary, table



"""Make pair plot of parameters"""

from tarpan.shared.info_path import InfoPath
from tarpan.shared.param_names import filter_param_names
from tarpan.shared.pair_plot import (
    PairPlotParams, save_pair_plot as shared_save_pair_plot)


def save_pair_plot(fit,
                   param_names=None,
                   info_path=InfoPath(),
                   pair_plot_params=PairPlotParams()):
    """
    Save a pair plot of distributions of parameters. It helps
    to see correlations between parameters and spot funnel
    shaped distributions that can result in sampling problems.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC
        Samples from cmdstanpy.

    param_names : list of str
        Names of parameters. Include all if None.

    info_path : InfoPath
        Path information for creating summaries.

    """

    info_path.set_codefile()
    param_names = filter_param_names(fit.column_names, param_names)
    samples = fit.get_drawset(params=param_names)

    shared_save_pair_plot(samples, param_names=param_names,
                          info_path=info_path,
                          pair_plot_params=pair_plot_params)



"""Manipulate names of model parameters"""

import re


"""Non-parameter columns returned by Stan"""
STAN_TECHNICAL_COLUMNS = [
    'lp__',
    'accept_stat__',
    'stepsize__',
    'treedepth__',
    'n_leapfrog__',
    'divergent__',
    'energy__']


def filter_param_names(stan_params, param_names=None):
    """
    Filter names of parmeters from `stan_params`
    based on names in `param_names`. The purpose of this function
    is to return numbered names like 'a.1', 'a.2', 'a.3' by just
    supplying 'a' in `param_names`.
    In addition, removes Stan's technical columns like 'stepsize__'.

    Parameters
    ---------

    stan_params: list of str
        Parameter names from stan.
        This can include names like 'a.1', 'a.2' etc.
    param_names: list of str
        Parameter names we want. It can include both numbered names
        ('a.1', 'a.2', 'a.3') and non-numbered ('a', 'b'). Return everything
        is None.

    Returns
    -------
    list of str:
        Filtered parameter names
    """

    # Remove technical columns
    stan_params = [
        a for a in stan_params if a not in STAN_TECHNICAL_COLUMNS]

    if param_names is not None:
        # If param_names contains 'a', we want both 'a' and
        # numbered parameter names 'a.1', 'a.2' etc.
        stan_params = [
            a for a in stan_params
            if a in param_names
            or (re.sub(r'\.[0-9]+\Z', '', a) in param_names)
        ]

    return stan_params



"""Contains main function that does all analysis"""

from tarpan.shared.info_path import InfoPath, get_info_path
from tarpan.shared.summary import SummaryParams
from tarpan.shared.tree_plot import make_tree_plot
from tarpan.shared.histogram import save_histogram
from tarpan.shared.pair_plot import save_pair_plot
from tarpan.cmdstanpy.summary import save_summary
from tarpan.cmdstanpy.traceplot import save_traceplot


def save_diagnostic(fit, info_path=InfoPath()):
    """
    Save diagnostic information from the fit into a text file.
    """

    info_path = InfoPath(**info_path.__dict__)
    info_path.base_name = info_path.base_name or 'diagnostic'
    info_path.extension = 'txt'
    file_path = get_info_path(info_path)

    with open(file_path, "w") as text_file:
        print(fit.diagnose(), file=text_file)


def save_analysis(fit, param_names=None, info_path=InfoPath(),
                  summary_params=SummaryParams()):
    """
    This is the most useful function of Tarpan library.

    It creates all analysis files: diagnostic, summary, trace and posterior.

    Parameters
    -----------

    fit : cmdstanpy.stanfit.CmdStanMCMC

        Contains the samples from cmdstanpy.

    param_names : list of str

        Names of parameters to plot. Plot all parameters if None.
    """

    info_path.set_codefile()
    save_diagnostic(fit, info_path=info_path)

    summary = save_summary(
        fit, param_names=param_names, info_path=info_path,
        summary_params=summary_params)

    make_tree_plot(summary['df'], param_names=param_names, info_path=info_path,
                   summary_params=summary_params)

    save_traceplot(fit, param_names=param_names, info_path=info_path)

    save_histogram(
        summary['samples'], summary['df'], param_names=param_names,
        info_path=info_path, summary_params=summary_params)

    save_pair_plot(
        summary['samples'], param_names=param_names,
        info_path=info_path)



"""Make histograms of posterior parameter distributions"""

from dataclasses import dataclass
import numpy as np
import matplotlib.pyplot as plt
import math
import seaborn as sns
from tarpan.shared.info_path import InfoPath, get_info_path
from tarpan.shared.summary import SummaryParams
from tarpan.shared.param_names import filter_param_names


@dataclass
class HistogramParams:
    title: str = None  # Plot's title
    max_plot_pages: int = 4  # Maximum number of plots to generate.
    num_plot_rows: int = 4  # Number of rows (subplots) in a plot.
    ncols: int = 2  # Number of columns in the plot.

    hist_color = "#00a6ff"  # Fill color for histogram bars
    hist_edge_color = "#FFFFFF"  # Edge color for the histogram bars

    # Colors and line styles for KDE lines of the error bars (HPDIs)
    # Sorted from largerst smallest HPDI values
    kde_colors = ['#FF9922', '#6666FF', '#44FF55']
    kde_line_styles = ['dotted', 'solid', '-.']


def save_histogram(samples, summary, param_names=None,
                   info_path=InfoPath(),
                   histogram_params=HistogramParams(),
                   summary_params=SummaryParams()):
    """
    Make histograms for the parameters from posterior destribution.

    Parameters
    -----------

    samples : Panda's DataFrame

        Each column contains samples from posterior distribution.

    summary : Panda's DataFrame

        Summary information about each column.

    param_names : list of str

        Names of the parameters for plotting. If None, all will be plotted.
    """

    info_path = InfoPath(**info_path.__dict__)

    figures_and_axes = make_histograms(
        samples, summary, param_names=param_names,
        params=histogram_params,
        summary_params=summary_params)

    base_name = info_path.base_name or "histogram"
    info_path.extension = info_path.extension or 'pdf'

    for i, figure_and_axis in enumerate(figures_and_axes):
        info_path.base_name = f'{base_name}_{i + 1:02d}'
        plot_path = get_info_path(info_path)
        fig = figure_and_axis[0]
        fig.savefig(plot_path, dpi=info_path.dpi)
        plt.close(fig)


def make_histogram_one_page(i_start, samples, summary, param_names,
                            params: HistogramParams,
                            summary_params=SummaryParams()):
    """
    Make a single file with histograms for the parameters
    from posterior destribution.
    """

    nrows = math.ceil((len(param_names) - i_start) / params.ncols)

    if nrows > params.num_plot_rows:
        nrows = params.num_plot_rows

    fig_height = 4 * nrows

    fig, ax = plt.subplots(
        nrows=nrows,
        ncols=params.ncols, figsize=(12, fig_height),
        squeeze=False)

    axes = ax.flatten()

    for i_axis, ax in enumerate(axes):
        i_param = i_start + i_axis

        if i_param >= len(param_names):
            break

        parameter = param_names[i_param]
        param_samples = samples[parameter]
        data = summary.loc[parameter]

        # Exclude extreme outliers from the samples
        # to avoid the blow-up of the x-axis range
        inner_range = np.percentile(param_samples, [0.5, 99.5])

        samples_for_kde = param_samples[(param_samples > inner_range[0])
                                        & (param_samples < inner_range[1])]

        sns.distplot(samples_for_kde, kde=False, norm_hist=True, ax=ax,
                     hist_kws={
                        "color": params.hist_color,
                        "zorder": 1,
                        "edgecolor": params.hist_edge_color,
                        "linewidth": 1,
                        "alpha": 1})

        # Show KDEs for the error bars (HPDIs)
        # -----------

        hpdis = sorted(summary_params.hpdi_percent(), reverse=True)

        for i, hpdi in enumerate(hpdis):
            start = f'{hpdi}CI-'
            end = f'{hpdi}CI+'

            # KDE plot
            sns.kdeplot(samples_for_kde, shade=False,
                        clip=[data[start], data[end]],
                        label=f'{hpdi}% HPDI', ax=ax, legend=None,
                        color=params.kde_colors[i],
                        linestyle=params.kde_line_styles[i],
                        linewidth=2)

            if i == len(hpdis) - 1:
                # Show shade under KDE for the last error bar
                sns.kdeplot(samples_for_kde, shade=True,
                            clip=[data[start], data[end]],
                            color="#000000",
                            label='_nolegend_', alpha=0.2,
                            zorder=10,
                            ax=ax, legend=None,
                            linewidth=2)

        ax.axvline(x=data['Mean'], label='Mean', linewidth=1.5,
                   linestyle='dashed', color='#33AA66')

        ax.axvline(x=data['Mode'], label='Mode', linewidth=1.5,
                   color='#FF66AA')

        ax.set_xlabel(parameter)

    # Do not draw the axes for non-existing plots
    for ax in axes[len(param_names):]:
        ax.axis('off')

    handles, labels = axes[0].get_legend_handles_labels()
    fig.legend(handles, labels, loc='upper center', mode='expand',
               ncol=len(labels))

    fig.tight_layout(rect=[0, 0.03, 1, 0.95])

    return (fig, ax)


def make_histograms(samples, summary, param_names=None,
                    params=HistogramParams(),
                    summary_params=SummaryParams()):
    """
    Make multiple files with
    histograms for the parameters from posterior destribution.

    Parameters
    -----------

    samples : Panda's DataFrame

        Each column contains samples from posterior distribution.

    summary : Panda's DataFrame

        Summary information about each column.

    param_names : list of str

        Names of the parameters for plotting. If None, all will be plotted.
    """
    param_names = filter_param_names(samples.columns, param_names)

    # Total number of plots
    n_plots = math.ceil(math.ceil(len(param_names) / params.ncols) /
                        params.num_plot_rows)

    if n_plots > params.max_plot_pages:
        print((
            f'Showing only first {params.max_plot_pages} '
            f'pages out of {n_plots} of histogram.'
            'Consider specifying "param_names".'))

        n_plots = params.max_plot_pages

    if n_plots < 1:
        n_plots = 1

    figures_and_axes = []

    # Make multiple traceplots
    for i_plot in range(n_plots):
        fig, ax = make_histogram_one_page(
            i_start=i_plot * params.num_plot_rows * params.ncols,
            samples=samples,
            summary=summary,
            param_names=param_names,
            params=params,
            summary_params=summary_params)

        figures_and_axes.append([fig, ax])

    return figures_and_axes



import shutil
import os

from tarpan.testutils.a01_eight_schools.eight_schools import (
    get_fit, get_fit_larger_uncertainties)

from tarpan.cmdstanpy.tree_plot import save_tree_plot
from tarpan.shared.tree_plot import TreePlotParams


def test_save_tree_plot():
    fit = get_fit()
    fit2 = get_fit_larger_uncertainties()
    fits = [fit, fit2]

    values_no_error_bars = [
        {
            "mu": 1.1,
            "tau": 3.1,
        }
    ]

    tree_params = TreePlotParams()

    tree_params.labels = [
        "Normal",
        "Larger uncertainties",
        "Exact",
    ]

    outdir = "tarpan/cmdstanpy/model_info/tree_plot_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_tree_plot(
        fits, extra_values=values_no_error_bars, param_names=["mu", "tau"],
        tree_params=tree_params)

    assert os.path.isfile(os.path.join(outdir, "summary.pdf"))



from tarpan.shared.param_names import filter_param_names


def test_filter_param_names():
    result = filter_param_names(['a', 'b', 'c'], ['a', 'b'])

    assert result == ['a', 'b']


def test_filter_param_names__numbered():
    result = filter_param_names(['a.1', 'a.2', 'a', 'b', 'c'], ['a', 'b'])

    assert result == ['a.1', 'a.2', 'a', 'b']


def test_filter_param_names__no_filter():
    result = filter_param_names(['a', 'b', 'c'])

    assert result == ['a', 'b', 'c']


def test_filter_param_names__remove_technical_columns():
    result = filter_param_names(['a', 'stepsize__', 'c'])

    assert result == ['a', 'c']



"""Make traceplots from Stan's output samples"""

from dataclasses import dataclass
import math
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from tarpan.shared.info_path import InfoPath, get_info_path
from tarpan.plot.utils import plot_kde_fallback_hist, remove_ticks_labels
from tarpan.shared.param_names import filter_param_names


@dataclass
class TraceplotParams:
    max_traceplot_pages = 4  # Maximum number of traceplot to generate.
    num_traceplot_rows = 8  # Number of rows (subplots) in a traceplot.

    # Maximum number of samples to show in one traceplot.
    show_max_samples = 300

    color = ['#00993399', '#FF992299', '#8888FF99',
             '#BBBB1199', '#3399AA99', '#33BB1199',
             '#33007799', '#BBAA2299', '#55FFBB99']

    color_samples = ['#00993350', '#FF992250', '#8888FF50',
                     '#BBBB1150', '#3399AA50', '#33BB1150',
                     '#33007750', '#BBAA2250', '#55FFBB50']


def save_traceplot(fit, param_names=None, info_path=InfoPath(),
                   traceplot_params=TraceplotParams()):
    """
    Saves traceplots form the fit.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC

        Contains the samples from cmdstanpy.

    param_names : list of str

        Names of parameters to plot.
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)

    figures_and_axes = traceplot(
        fit, param_names=param_names, params=traceplot_params)

    base_name = info_path.base_name or "traceplot"
    info_path.extension = info_path.extension or 'pdf'

    for i, figure_and_axis in enumerate(figures_and_axes):
        info_path.base_name = f'{base_name}_{i + 1:02d}'
        plot_path = get_info_path(info_path)
        fig = figure_and_axis[0]
        fig.savefig(plot_path, dpi=info_path.dpi)
        plt.close(fig)


def make_single_traceplot(i_start, fit,
                          param_names,
                          params: TraceplotParams):
    """
    Show one page of traceplots.

    Parameters
    ----------

    param_names : list of str

        Names of parameters to plot.

    """

    nrows = fit.sample.shape[2] - i_start

    if nrows > params.num_traceplot_rows:
        nrows = params.num_traceplot_rows

    if nrows > len(param_names) - i_start:
        nrows = len(param_names) - i_start

    fig_height = 2 * nrows

    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=2, figsize=(12, fig_height),
        squeeze=False)

    # If there are too many samples, plotting will be slow
    # Thus, we only use no more than `show_max_samples`
    # by taking only every `draw_every` sample
    draw_every = int(fit.draws / params.show_max_samples)

    if draw_every < 1:
        draw_every = 1

    for i_row in range(nrows):
        param_name = param_names[i_start + i_row]
        # Parameter index in sample array
        i_param = fit.column_names.index(param_name)

        for i_chain in range(fit.chains):
            samples_for_kde = fit.sample[:, i_chain, i_param]

            # Plot KDEs
            # ----------

            ax = axes[i_row, 0]

            # Exclude extreme outliers from the samples
            # to avoid the blow-up of the x-axis range

            inner_range = np.percentile(samples_for_kde, [0.5, 99.5])

            samples_for_kde = samples_for_kde[
                (samples_for_kde > inner_range[0])
                & (samples_for_kde < inner_range[1])]

            plot_kde_fallback_hist(samples_for_kde, ax=ax,
                                   color=params.color[i_chain])

            remove_ticks_labels(ax, remove_x=False, remove_y=True)
            ax.set_ylabel(param_name)

            # Plot samples
            # ----------

            samples = fit.sample[::draw_every, i_chain, i_param]
            ax = axes[i_row, 1]
            ax.plot(samples, color=params.color_samples[i_chain])
            remove_ticks_labels(ax)

    fig.tight_layout()

    return (fig, axes)


def traceplot(fit, param_names=None, params=TraceplotParams()):
    """
    Show traceplots, diagnostic plots of samples for all
    parameters for all chains.

    Parameters
    ----------

    param_names: list of str

        List of parameters to plot.  If None, plot all.

    """

    sns.set(style="ticks")

    # Make the list of columns to be shown in the plots
    param_names = filter_param_names(fit.column_names, param_names)
    param_names.insert(0, 'lp__')  # Always show traceplot of probability

    # Total number of plots
    n_plots = math.ceil(len(param_names) / params.num_traceplot_rows)

    if n_plots > params.max_traceplot_pages:
        print((
            f'Traceplot shows only first {params.max_traceplot_pages} '
            f'pages out of {n_plots}. Consider specifying "param_names".'))

        n_plots = params.max_traceplot_pages

    if n_plots < 1:
        n_plots = 1

    figures_and_axes = []

    # Make multople traceplots
    for i_plot in range(n_plots):
        fig, ax = make_single_traceplot(
            i_start=i_plot * params.num_traceplot_rows,
            fit=fit,
            param_names=param_names,
            params=params)

        figures_and_axes.append([fig, ax])

    return figures_and_axes
